pub struct Locks { /* private fields */ }Expand description
Avoid and track conflicts between packages.
Implementations
Sourceimpl Locks
impl Locks
Sourcepub fn new(preview: Preview) -> Self
pub fn new(preview: Preview) -> Self
Create a new Locks instance with the given preview settings.
Sourcepub fn warn_package_conflicts(self) -> Result<(), Error>
pub fn warn_package_conflicts(self) -> Result<(), Error>
Warn when the same file with different contents exists in multiple packages.
The intent is to detect different variants of the same package installed over each other, or different packages using the same top-level module name, which cause non-deterministic failures only surfacing at runtime. See https://github.com/astral-sh/uv/pull/13437 for a list of cases.
The check has some false negatives. It is rather too lenient than too strict, and includes
support for namespace packages that include the same __init__.py file, e.g., gpu-a and
gpu-b both including the same gpu/__init__.py.
We assume that all wheels of a package have the same module(s), so a conflict between installing two unpacked wheels is a conflict between two packages.
Performance
When there are no namespace packages, this method is a Mutex lock and a hash map iteration.
When there are namespace packages, we only traverse into directories shared by at least two
packages. For example, for namespace packages gpu-a, gpu-b, and gpu-c with
gpu/a/__init__.py, gpu/b/__init__.py, and gpu/c/__init__.py respectively, we only
need to read the gpu directory. If there is a deeper shared directory, we only recurse
down to this directory. As packages without conflicts generally do not share many
directories, we do not recurse far.
For each directory, we analyze all packages sharing the directory at the same time, reading the directory in each unpacked wheel only once. Effectively, we perform a parallel directory walk with early exit.
We avoid reading the actual file contents and assume they are the same when their file
length matches. This also excludes the same empty __init__.py files being reported as
conflicting.